@@ -1,7 +1,9 @@
 # -*- coding: utf-8 -*-
+"""Back-port compiler for Python 3.8 assignment expressions."""
 
 import argparse
 import glob
+import io
 import locale
 import os
 import re
@@ -31,7 +33,7 @@ finally:    # alias and aftermath
     del multiprocessing
 
 # version string
-__version__ = '0.1.1'
+__version__ = '0.1.2'
 
 # from configparser
 BOOLEAN_STATES = {'1': True, '0': False,
@@ -77,10 +79,51 @@ def predicate(filename):  # pragma: no cover
 
 tbtrim.set_trim_rule(predicate, strict=True, target=(ConvertError, ContextError))
 
+###############################################################################
+# UUID 4 generator wrapper
+
+
+class UUID4Generator:
+    """UUID 4 generator wrapper to prevent UUID collisions."""
+
+    def __init__(self, dash=True):
+        """Constructor of UUID 4 generator wrapper.
+
+        Args:
+         - `dash` -- `bool`, whether the generated UUID string has dashes or not
+
+        """
+        self.used_uuids = set()
+        self.dash = dash
+
+    def gen(self):
+        """Generate a new UUID 4 string that is guaranteed not to collide with used UUIDs.
+
+        Returns:
+         - `str` -- a new UUID 4 string
+
+        """
+        while True:
+            nuid = uuid.uuid4()
+            nuid = str(nuid) if self.dash else nuid.hex
+            if nuid not in self.used_uuids:
+                break
+        self.used_uuids.add(nuid)
+        return nuid
+
+
+# Initialise a UUID4Generator for filename UUIDs.
+# In the conversion process, this will be replaced by a new UUID4Generator for identifier UUIDs.
+uuid_gen = UUID4Generator(dash=True)
+
 ###############################################################################
 # Main convertion implementation
 
 # walrus wrapper template
+NAME_TEMPLATE = '''\
+if False:
+%(tabsize)s%(name_list)s = NotImplemented
+'''.splitlines()  # `str.splitlines` will remove trailing newline
 CALL_TEMPLATE = '__walrus_wrapper_%(name)s_%(uuid)s(%(expr)s)'
 FUNC_TEMPLATE = '''\
 def __walrus_wrapper_%(name)s_%(uuid)s(expr):
@@ -90,11 +133,19 @@ def __walrus_wrapper_%(name)s_%(uuid)s(expr):
 %(tabsize)sreturn %(name)s
 '''.splitlines()  # `str.splitlines` will remove trailing newline
 
+# special template for lambda
+LAMBDA_CALL_TEMPLATE = '__walrus_wrapper_lambda_%(uuid)s'
+LAMBDA_FUNC_TEMPLATE = '''\
+def __walrus_wrapper_lambda_%(uuid)s(%(param)s):
+%(tabsize)s"""Wrapper function for lambda definitions."""
+%(tabsize)s%(suite)s
+'''.splitlines()  # `str.splitlines` will remove trailing newline
+
 # special templates for ClassVar
 ## locals dict
 LCL_DICT_TEMPLATE = '_walrus_wrapper_%(cls)s_dict = dict()'
 LCL_NAME_TEMPLATE = '_walrus_wrapper_%(cls)s_dict[%(name)r]'
-LCL_CALL_TEMPLATE = '__WalrusWrapper%(cls)s.get_%(name)s_%(uuid)s(locals())'
+LCL_CALL_TEMPLATE = '__WalrusWrapper%(cls)s.get_%(name)s_%(uuid)s()'
 LCL_VARS_TEMPLATE = '''\
 [setattr(%(cls)s, k, v) for k, v in _walrus_wrapper_%(cls)s_dict.items()]
 del _walrus_wrapper_%(cls)s_dict
@@ -113,19 +164,11 @@ CLS_FUNC_TEMPLATE = '''\
 %(tabsize)s%(tabsize)sreturn _walrus_wrapper_%(cls)s_dict[%(name)r]
 
 %(tabsize)s@staticmethod
-%(tabsize)sdef get_%(name)s_%(uuid)s(locals_=locals()):
+%(tabsize)sdef get_%(name)s_%(uuid)s():
 %(tabsize)s%(tabsize)s"""Wrapper function for assignment expression."""
-%(tabsize)s%(tabsize)s# get value from buffer dict
-%(tabsize)s%(tabsize)stry:
+%(tabsize)s%(tabsize)sif %(name)r in _walrus_wrapper_%(cls)s_dict:
 %(tabsize)s%(tabsize)s%(tabsize)sreturn _walrus_wrapper_%(cls)s_dict[%(name)r]
-%(tabsize)s%(tabsize)sexcept KeyError:
-%(tabsize)s%(tabsize)s%(tabsize)spass
-
-%(tabsize)s%(tabsize)s# get value from locals dict
-%(tabsize)s%(tabsize)stry:
-%(tabsize)s%(tabsize)s%(tabsize)sreturn locals_[%(name)r]
-%(tabsize)s%(tabsize)sexcept KeyError as error:
-%(tabsize)s%(tabsize)s%(tabsize)sraise NameError('name %%r is not defined' %% %(name)r).with_traceback(error.__traceback__)
+%(tabsize)s%(tabsize)sraise NameError('name %%r is not defined' %% %(name)r)
 '''.splitlines()  # `str.splitlines` will remove trailing newline
 
 
@@ -163,6 +206,10 @@ class Context:
     def string(self):
         return self._buffer
 
+    @property
+    def lambdef(self):
+        return self._lamb
+
     @property
     def variables(self):
         return self._vars
@@ -223,6 +270,7 @@ class Context:
         self._buffer = ''  # final result
 
         self._vars = list()  # variable initialisation
+        self._lamb = list()  # converted lambda definitions ({param, suite, uuid})
         self._func = list()  # wrapper functions ({name, uuid, keyword})
 
         self._walk(node)  # traverse children
@@ -286,12 +334,13 @@ class Context:
         # leaf node
         self += node.get_code()
 
-    def _process_suite_node(self, node, func=False, cls_ctx=None):
+    def _process_suite_node(self, node, func=False, raw=False, cls_ctx=None):
         """Process indented suite (`suite` or ...).
 
         Args:
          - `node` -- `Union[parso.python.tree.PythonNode, parso.python.tree.PythonLeaf]`, suite node
          - `func` -- `bool`, if the suite is of function definition
+         - `raw` -- `bool`, raw processing flag
          - `cls_ctx` -- `Optional[str]`, class name when suite if of class contextion
 
         """
@@ -311,14 +360,19 @@ class Context:
         if cls_ctx is None:
             ctx = Context(node=node, context=self._context,
                           column=indent, tabsize=self._tabsize,
-                          linesep=self._linesep, keyword=keyword)
+                          linesep=self._linesep, keyword=keyword, raw=raw)
         else:
             ctx = ClassContext(cls_ctx=cls_ctx,
                                node=node, context=self._context,
                                column=indent, tabsize=self._tabsize,
-                               linesep=self._linesep, keyword=keyword)
-
+                               linesep=self._linesep, keyword=keyword, raw=raw)
         self += ctx.string.lstrip()
+
+        # keep records
+        if raw:
+            self._lamb.extend(ctx.lambdef)
+            self._vars.extend(ctx.variables)
+            self._func.extend(ctx.functions)
         self._context.extend(ctx.global_stmt)
 
     def _process_namedexpr_test(self, node):
@@ -331,7 +385,7 @@ class Context:
         # split assignment expression
         node_name, _, node_expr = node.children
         name = node_name.value
-        nuid = uuid.uuid4().hex
+        nuid = uuid_gen.gen()
 
         # calculate expression string
         ctx = Context(node=node_expr, context=self._context,
@@ -472,6 +526,44 @@ class Context:
         # suite
         self._process_suite_node(func_suite, func=True)
 
+    def _process_lambdef(self, node):
+        """Process lambda definition (``lambdef``).
+
+        Args:
+         - `node` -- `parso.python.tree.Lambda`, lambda node
+
+        """
+        if not self.has_walrus(node):
+            self += node.get_code()
+            return
+
+        children = iter(node.children)
+
+        # <Keyword: lambda>
+        next(children)
+
+        # vararglist
+        para_list = list()
+        for child in children:
+            if child.type == 'operator' and child.value == ':':
+                break
+            para_list.append(child)
+        param = ''.join(map(lambda n: n.get_code(), para_list))
+
+        # test_nocond | test
+        indent = self._column + self._tabsize
+        ctx = LambdaContext(node=next(children), context=self._context,
+                            column=indent, tabsize=self._tabsize,
+                            linesep=self._linesep, keyword='nonlocal')
+        suite = ctx.string.strip()
+
+        # keep record
+        nuid = uuid_gen.gen()
+        self._lamb.append(dict(param=param, suite=suite, uuid=nuid))
+
+        # replacing lambda
+        self += LAMBDA_CALL_TEMPLATE % dict(uuid=nuid)
+
     def _process_if_stmt(self, node):
         """Process if statement (``if_stmt``).
 
@@ -660,7 +752,7 @@ class Context:
 
         # first, the prefix codes
         self._buffer += self._prefix + prefix
-        if flag and self._linting and self._vars:
+        if flag and self._linting and self._buffer:
             if (self._node_before_walrus is not None \
                     and self._node_before_walrus.type in ('funcdef', 'classdef') \
                     and self._column == 0):
@@ -677,27 +769,33 @@ class Context:
             linesep = self._linesep * (1 if self._column > 0 else 2)
         else:
             linesep = ''
-        for var in sorted(set(self._vars)):
-            self._buffer += '%(indent)s%(name)s = locals().get(%(name)r)%(linesep)s' % dict(
-                indent=indent, name=var, linesep=self._linesep,
-            )
+        if self._vars:
+            name_list = ' = '.join(sorted(set(self._vars)))
+            self._buffer += indent + (
+                '%s%s' % (self._linesep, indent)
+            ).join(NAME_TEMPLATE) % dict(tabsize=tabsize, name_list=name_list) + self._linesep
         for func in sorted(self._func, key=lambda func: func['name']):
-            self._buffer += linesep + indent + (
+            if self._buffer:
+                self._buffer += linesep
+            self._buffer += indent + (
                 '%s%s' % (self._linesep, indent)
             ).join(FUNC_TEMPLATE) % dict(tabsize=tabsize, **func) + self._linesep
+        for lamb in self._lamb:
+            if self._buffer:
+                self._buffer += linesep
+            self._buffer += indent + (
+                '%s%s' % (self._linesep, indent)
+            ).join(LAMBDA_FUNC_TEMPLATE) % dict(tabsize=tabsize, **lamb) + self._linesep
 
         # finally, the suffix codes
-        if flag and self._linting and self._vars:
+        if flag and self._linting:
             blank = 2 if self._column == 0 else 1
             self._buffer += self._linesep * self.missing_whitespaces(prefix=self._buffer, suffix=suffix,
                                                                      blank=blank, linesep=self._linesep)
         self._buffer += suffix
 
     def _strip(self):
-        """Strip comments from string.
-
-        Args:
-         - `string` -- `str`, buffer string
+        """Strip comments from suffix buffer.
 
         Returns:
          - `str` -- prefix comments
@@ -707,7 +805,7 @@ class Context:
         prefix = ''
         suffix = ''
 
-        lines = iter(self._suffix.splitlines(True))
+        lines = io.StringIO(self._suffix, newline=self._linesep)
         for line in lines:
             if line.strip().startswith('#'):
                 prefix += line
@@ -858,12 +956,14 @@ class Context:
          - `int` -- number of preceding blank lines
 
         """
-        count = -1  # keep trailing newline in `prefix`
+        count = 0
         if prefix:
             for line in reversed(prefix.split(linesep)):
                 if line.strip():
                     break
                 count += 1
+            if count > 0:  # keep trailing newline in `prefix`
+                count -= 1
         if suffix:
             for line in suffix.split(linesep):
                 if line.strip():
@@ -873,7 +973,6 @@ class Context:
         if count < 0:
             count = 0
         missing = blank - count
-
         if missing > 0:
             return missing
         return 0
@@ -909,6 +1008,49 @@ class Context:
         return prefix, suffix
 
 
+class LambdaContext(Context):
+    """Lambda (suite) conversion context."""
+
+    def _concat(self):
+        """Concatenate final string."""
+        flag = self.has_walrus(self._root)
+
+        # first, the variables and functions
+        indent = '\t'.expandtabs(self._column)
+        tabsize = '\t'.expandtabs(self._tabsize)
+        if self._linting:
+            linesep = self._linesep * (1 if self._column > 0 else 2)
+        else:
+            linesep = ''
+        if self._vars:
+            name_list = ' = '.join(sorted(set(self._vars)))
+            self._buffer += indent + (
+                '%s%s' % (self._linesep, indent)
+            ).join(NAME_TEMPLATE) % dict(tabsize=tabsize, name_list=name_list) + self._linesep
+        for func in sorted(self._func, key=lambda func: func['name']):
+            if self._buffer:
+                self._buffer += linesep
+            self._buffer += indent + (
+                '%s%s' % (self._linesep, indent)
+            ).join(FUNC_TEMPLATE) % dict(tabsize=tabsize, **func) + self._linesep
+        for lamb in self._lamb:
+            if self._buffer:
+                self._buffer += linesep
+            self._buffer += indent + (
+                '%s%s' % (self._linesep, indent)
+            ).join(LAMBDA_FUNC_TEMPLATE) % dict(tabsize=tabsize, **lamb) + self._linesep
+        if flag and self._linting:
+            blank = 2 if self._column == 0 else 1
+            self._buffer += self._linesep * self.missing_whitespaces(prefix=self._buffer, suffix=self._prefix,
+                                                                     blank=blank, linesep=self._linesep)
+
+        # then, the `return` statement
+        self._buffer += indent + 'return'
+
+        # finally, the source codes
+        self._buffer += self._prefix + self._suffix
+
+
 class ClassContext(Context):
     """Class (suite) conversion context."""
 
@@ -916,22 +1058,23 @@ class ClassContext(Context):
     def cls_var(self):
         return self._cls_var
 
-    def __init__(self, cls_ctx, node,
+    def __init__(self, node,
+                 cls_ctx, cls_var=None,
                  column=0, tabsize=None,
                  linesep=None, keyword=None,
-                 context=None, raw=False, cls_var=None):
+                 context=None, raw=False):
         """Conversion context.
 
         Args:
-         - `cls_ctx` -- `str`, class context name
          - `node` -- `Union[parso.python.tree.PythonNode, parso.python.tree.PythonLeaf]`, parso AST
+         - `cls_ctx` -- `str`, class context name
+         - `cls_var` -- `Dict[str, str]`, mapping for assignment variable and its UUID
          - `column` -- `int`, current indentation level
          - `tabsize` -- `Optional[int]`, indentation tab size
          - `linesep` -- `Optional[str]`, line seperator
          - `keyword` -- `Optional[str]`, keyword for wrapper function
          - `context` -- `Optional[List[str]]`, global context
          - `raw` -- `bool`, raw context processing flag
-         - `cls_var` -- `Dict[str, str]`, mapping for assignment variable and its UUID
 
         Envs:
          - `WALRUS_LINESEP` -- line separator to process source files (same as `--linesep` option in CLI)
@@ -950,12 +1093,13 @@ class ClassContext(Context):
                          column=column, tabsize=tabsize,
                          linesep=linesep, keyword=keyword, raw=raw)
 
-    def _process_suite_node(self, node, func=False, cls_ctx=None):
+    def _process_suite_node(self, node, func=False, raw=False, cls_ctx=None):
         """Process indented suite (`suite` or ...).
 
         Args:
          - `node` -- `Union[parso.python.tree.PythonNode, parso.python.tree.PythonLeaf]`, suite node
          - `func` -- `bool`, if the suite is of function definition
+         - `raw` -- `bool`, raw context processing flag
          - `cls_ctx` -- `Optional[str]`, class name when suite if of class contextion
 
         """
@@ -985,8 +1129,14 @@ class ClassContext(Context):
                                node=node, context=self._context,
                                column=indent, tabsize=self._tabsize,
                                linesep=self._linesep, keyword=keyword)
-
         self += ctx.string.lstrip()
+
+        # keep record
+        if raw:
+            self._lamb.extend(ctx.lambdef)
+            self._vars.extend(ctx.variables)
+            self._func.extend(ctx.functions)
+            self._cls_var.update(ctx.cls_var)
         self._context.extend(ctx.global_stmt)
 
     def _process_namedexpr_test(self, node):
@@ -999,7 +1149,7 @@ class ClassContext(Context):
         # split assignment expression
         node_name, _, node_expr = node.children
         name = node_name.value
-        nuid = uuid.uuid4().hex
+        nuid = uuid_gen.gen()
 
         # calculate expression string
         ctx = ClassContext(cls_ctx=self._cls_ctx, cls_var=self._cls_var,
@@ -1007,6 +1157,7 @@ class ClassContext(Context):
                            column=self._column, tabsize=self._tabsize,
                            linesep=self._linesep, keyword=self._keyword, raw=True)
         expr = ctx.string.strip()
+        self._lamb.extend(ctx.lambdef)
         self._vars.extend(ctx.variables)
         self._func.extend(ctx.functions)
         self._cls_var.update(ctx.cls_var)
@@ -1035,7 +1186,7 @@ class ClassContext(Context):
 
         """
         name = node.value
-        nuid = uuid.uuid4().hex
+        nuid = uuid_gen.gen()
 
         prefix, _ = self.extract_whitespaces(node)
         self += prefix + LCL_NAME_TEMPLATE % dict(cls=self._cls_ctx, name=name)
@@ -1108,7 +1259,9 @@ class ClassContext(Context):
                 '%s%s' % (self._linesep, indent)
             ).join(CLS_NAME_TEMPLATE) % dict(tabsize=tabsize, cls=self._cls_ctx) + linesep
         for func in sorted(self._func, key=lambda func: func['name']):
-            self._buffer += linesep + indent + (
+            if self._buffer:
+                self._buffer += linesep
+            self._buffer += indent + (
                 '%s%s' % (self._linesep, indent)
             ).join(CLS_FUNC_TEMPLATE) % dict(tabsize=tabsize, cls=self._cls_ctx, **func) + linesep
 
@@ -1136,6 +1289,10 @@ def convert(string, source='<unknown>'):
      - `str` -- converted string
 
     """
+    # Initialise new UUID4Generator for identifier UUIDs
+    global uuid_gen
+    uuid_gen = UUID4Generator(dash=False)
+
     # parse source string
     module = parse(string, source)
 
@@ -1168,14 +1325,14 @@ def walrus(filename):
     encoding = os.getenv('WALRUS_ENCODING', LOCALE_ENCODING)
 
     # file content
-    with open(filename, 'r', encoding=encoding) as file:
+    with open(filename, 'r', encoding=encoding, newline='') as file:
         text = file.read()
 
     # do the dirty things
     text = convert(text, filename)
 
     # dump back to the file
-    with open(filename, 'w', encoding=encoding) as file:
+    with open(filename, 'w', encoding=encoding, newline='') as file:
         file.write(text)
 
 
@@ -1265,7 +1422,7 @@ def rename(path, root):
 
     """
     stem, ext = os.path.splitext(path)
-    name = '%s-%s%s' % (stem, uuid.uuid4(), ext)
+    name = '%s-%s%s' % (stem, uuid_gen.gen(), ext)
     return os.path.join(root, name)
 
 
